<?php

namespace Tests\Feature;

use Carbon\Carbon;
use Tests\TestCase;
use Illuminate\Testing\TestResponse;
use App\domain\travelport\CreateOAuthTokens;

class TravelPortAPITest extends TestCase
{
    protected CreateOAuthTokens $createOAuthTokens;
    protected string $reservationId;

    public function setUp(): void
    {
        parent::setUp();
        $this->createOAuthTokens = new CreateOAuthTokens();
    }

    /** @test */
    public function when_request_oauth_token()
    {
        $token = $this->createOAuthTokens->requestOAuthAccessToken();

        $this->assertSeeAccessTokenAndRefreshTokenStoredInSession($token);
    }

    /** @test */
    public function when_refresh_token_has_and_request_new_access_token()
    {
        $this->createOAuthTokens->requestOAuthAccessToken();
        $token = $this->createOAuthTokens->refreshOAuthAccessToken();

        $this->assertSeeAccessTokenAndRefreshTokenStoredInSession($token);
    }

    /** @test */
    public function when_search_flight_offers_for_round_trip()
    {
        $departureDate = Carbon::now()->addMonth(1);
        $arrivalDate = $departureDate->clone()->addDays(7);

        $searchCriteria = [
            "passengerCriteria" => [
                [
                    "@type" => "PassengerCriteria",
                    "number" => 1,
                    "passengerTypeCode" => "ADT"
                ]
            ],
            'searchCriteriaFlight' => [
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $departureDate->toDateString(),
                    "From" => [
                        "value" => "LON"
                    ],
                    "To" => [
                        "value" => "CMB"
                    ]
                ],
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $arrivalDate->toDateString(),
                    "From" => [
                        "value" => "CMB"
                    ],
                    "To" => [
                        "value" => "LON"
                    ]
                ],
            ]
        ];

        $response = $this->postJson('api/search_flight_offers', $searchCriteria);

        $this->assertFlightOffersReceivedWithRequiredStructure($response);
    }

    /** @test */
    public function when_search_flight_offers_for_round_trip_only_business_cabin()
    {
        $response = $this->requestBusinessCabinFlightOffersForRoundTrip();

        $this->assertOnlyBusinessOffersReceived($response);
    }

    /** @test */
    public function when_search_flight_offers_non_stop_only()
    {
        $response = $this->requestNonStopFlightOffersForRoundTrip();

        $this->assertReceivedOnlyNonStopFlightOffers($response);
    }

    /** @test */
    public function when_request_price_for_selected_offer()
    {
        $response = $this->requestPriceForOneWayOffer();

        $this->assertReceivedCompletePricingDetailsForSelectedOffers($response);
    }

    /** @test */
    public function when_create_booking_workbench()
    {
        $response = $this->initiateWorkbench();

        $this->assertWorkbenchCreatedAndReservationIdReceived($response);
    }

    /** @test */
    public function when_add_offer_in_booking_workbench()
    {
        $response = $this->createReservationAndAddOfferToWorkbench();

        $this->assertOfferAddedToWorkbenchAndOfferIdReceived($response);
    }

    /** @test */
    public function when_add_traveler_in_workbench()
    {
        $response = $this->createReservationAddOfferAndAddTraveler();

        $this->assertTravelerAddedWorkbenchAndTravelerIdReceived($response);
    }

    /** @test */
    public function when_commit_reservation_at_end_of_workbench()
    {
        $this->createReservationAddOfferAndAddTraveler();

        $response = $this->postJson('api/workbench/commit', [
            'reservationId' => $this->reservationId
        ]);

        $this->assertReservationCommitted($response);
    }

    protected function createReservationAddOfferAndAddTraveler(): TestResponse
    {
        $this->createReservationAndAddOfferToWorkbench();

        $traveler = [
            "Traveler" => [
                "gender" => "Male",
                "birthDate" => "1987-11-20",
                "passengerTypeCode" => "ADT",
                "PersonName" => [
                    "@type" => "PersonNameDetail",
                    "Given" => "Carlton Test First",
                    "Surname" => "Carlton Test Last"
                ],
                "Telephone" => [
                    [
                        "@type" => "Telephone",
                        "countryAccessCode" => "1",
                        "phoneNumber" => "778951753",
                        "cityCode" => "ORD",
                        "role" => "Home"
                    ]
                ],
                "Email" => [
                    [
                        "value" => "carltontest20231012@gmail.com"
                    ]
                ],
                "TravelDocument" => [
                    [
                        "@type" => "TravelDocumentDetail",
                        "docNumber" => "A20231012",
                        "docType" => "Passport",
                        "expireDate" => "2033-01-31",
                        "issueCountry" => "US",
                        "birthDate" => "1987-11-20",
                        "Gender" => "Male",
                        "PersonName" => [
                            "@type" => "PersonName",
                            "Given" => "Carlton Test First",
                            "Surname" => "Carlton Test Last"
                        ]
                    ]
                ]
            ]
        ];

        return $this->postJson('api/workbench/add_traveler', [
            'reservationId' => $this->reservationId,
            'traveller' => $traveler,
        ]);
    }

    protected function createReservationAndAddOfferToWorkbench(): TestResponse
    {
        $selectedOfferPrice = $this->requestPriceForOneWayOffer();

        $this->reservationId = $this->initiateWorkbench()
            ->json('ReservationResponse.Reservation.Identifier.value');

        $catalogId = substr($selectedOfferPrice->json('OfferListResponse.Identifier.value'), 0, -3);
        $offerId = $selectedOfferPrice->json('OfferListResponse.OfferID.0.id');
        $productId = $selectedOfferPrice->json('OfferListResponse.OfferID.0.Product.0.id');

        $catalogProductOffer = [
            "reservationId" => $this->reservationId,
            "catalogId" => $catalogId,
            "products" => [
                [
                    "offerId" => $offerId,
                    "productId" => $productId,
                ]
            ]
        ];

        return $this->postJson('api/workbench/add_offer', $catalogProductOffer);
    }

    protected function requestBusinessCabinFlightOffersForRoundTrip(): TestResponse
    {
        $departureDate = Carbon::now()->addMonth(1);
        $arrivalDate = $departureDate->clone()->addDays(7);

        $searchCriteria = [
            "passengerCriteria" => [
                [
                    "@type" => "PassengerCriteria",
                    "number" => 1,
                    "passengerTypeCode" => "ADT"
                ]
            ],
            'searchCriteriaFlight' => [
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $departureDate->toDateString(),
                    "From" => [
                        "value" => "LON"
                    ],
                    "To" => [
                        "value" => "CMB"
                    ]
                ],
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $arrivalDate->toDateString(),
                    "From" => [
                        "value" => "CMB"
                    ],
                    "To" => [
                        "value" => "LON"
                    ]
                ],
            ],
            "CabinPreference" => [
                [
                    "@type" => "CabinPreference",
                    "preferenceType" => "Preferred",
                    "cabins" => [
                        "Business"
                    ]
                ]
            ]

        ];

        return $this->requestFlightOffers($searchCriteria);
    }

    protected function requestNonStopFlightOffersForRoundTrip(): TestResponse
    {
        $departureDate = Carbon::now()->addMonth(1);
        $arrivalDate = $departureDate->clone()->addDays(7);

        $searchCriteria = [
            "passengerCriteria" => [
                [
                    "@type" => "PassengerCriteria",
                    "number" => 1,
                    "passengerTypeCode" => "ADT"
                ]
            ],
            'searchCriteriaFlight' => [
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $departureDate->toDateString(),
                    "From" => [
                        "value" => "LON"
                    ],
                    "To" => [
                        "value" => "CMB"
                    ]
                ],
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $arrivalDate->toDateString(),
                    "From" => [
                        "value" => "CMB"
                    ],
                    "To" => [
                        "value" => "LON"
                    ]
                ],
            ],
            "flightType" => [
                "connectionType" => "NonStopDirect",
            ]

        ];

        return $this->requestFlightOffers($searchCriteria);
    }

    protected function requestPriceForOneWayOffer(): TestResponse
    {
        list($catalogProductOfferingsIdentifier, $offerId, $productId)
            = $this->getFirstProductOfferFromOneWayOffersList();

        $catalogProductOffer = [
            "catalogId" => $catalogProductOfferingsIdentifier,
            "products" => [
                [
                    "offerId" => $offerId,
                    "productId" => $productId,
                ]
            ]
        ];

        return $this->postJson('api/flight_offers_pricing', $catalogProductOffer);
    }

    protected function getFirstProductOfferFromOneWayOffersList(): array
    {
        $flightOffers = $this->requestFlightOffersForOneWay();

        $dynamicKey = array_key_first($flightOffers['productOffers']);

        $catalogProductOfferingsIdentifier = $flightOffers['productOffers'][$dynamicKey][0]['catalogProductOfferingsIdentifier'];
        $offerId = $flightOffers['productOffers'][$dynamicKey][0]['offerId'];
        $productId = $flightOffers['productOffers'][$dynamicKey][0]['id'];

        return array($catalogProductOfferingsIdentifier, $offerId, $productId);
    }

    protected function requestFlightOffersForOneWay(): TestResponse
    {
        $departureDate = Carbon::now()->addMonth(1);
        $arrivalDate = $departureDate->clone()->addDays(7);

        $searchCriteria = [
            "passengerCriteria" => [
                [
                    "@type" => "PassengerCriteria",
                    "number" => 1,
                    "passengerTypeCode" => "ADT"
                ]
            ],
            'searchCriteriaFlight' => [
                [
                    "@type" => "SearchCriteriaFlight",
                    "departureDate" => $departureDate->toDateString(),
                    "From" => [
                        "value" => "LON"
                    ],
                    "To" => [
                        "value" => "CMB"
                    ]
                ],
            ],
            "CabinPreference" => [
                [
                    "@type" => "CabinPreference",
                    "preferenceType" => "Preferred",
                    "cabins" => [
                        "Business"
                    ]
                ]
            ]

        ];

        return $this->requestFlightOffers($searchCriteria);
    }

    protected function initiateWorkbench(): TestResponse
    {
        return $this->getJson('api/workbench/create_reservation');
    }

    protected function requestFlightOffers(array $searchCriteria): TestResponse
    {
        return $this->postJson('api/search_flight_offers', $searchCriteria);
    }

    protected function assertSeeAccessTokenAndRefreshTokenStoredInSession(?array $token): void
    {
        $this->assertAccessTokenAndRefreshTokenReceived($token);
        $this->assertAccessTokenAndRefreshTokenStoredInSession();
    }

    protected function assertAccessTokenAndRefreshTokenReceived(?array $token): void
    {
        $this->assertArrayHasKey("access_token", $token);
        $this->assertArrayHasKey("refresh_token", $token);
    }

    protected function assertAccessTokenAndRefreshTokenStoredInSession(): void
    {
        $this->assertIsString(session('access_token'));
        $this->assertIsString(session('refresh_token'));
    }

    protected function assertFlightOffersReceivedWithRequiredStructure(TestResponse $response): void
    {
        $response->assertOk();
        $this->assertArrayHasKey('productOffers', $response->json());
        $response->assertJsonStructure([
            'productOffers' => [
                '*' => [
                    '*' => [
                        '@type',
                        'offerId',
                        'combinabilityCode',
                        'catalogProductOfferingsIdentifier',
                        'id',
                        'fromAirport',
                        'fromAirportCarrierCode',
                        'fromAirportAircraftCode',
                        'toAirport',
                        'toAirportCarrierCode',
                        'toAirportAircraftCode',
                        'originDepartureTime',
                        'destinationArrivalTime',
                        'duration',
                        'numberOfStops',
                        'numberOfBookableSeats',
                        'durationInHours',
                        'price',
                        'bestCombinablePrice',
                        'productAir',
                    ],
                ],
            ],
        ]);

    }

    protected function assertOnlyBusinessOffersReceived(TestResponse $response): void
    {
        $response->assertOk();
        $this->assertArrayHasKey('productOffers', $response->json());
        $response->assertJsonFragment(["cabin" => "Business"]);
        $response->assertJsonMissing(["cabin" => "Economy"]);
        $response->assertJsonMissing(["cabin" => "PremiumEconomy"]);
        $response->assertJsonMissing(["cabin" => "First"]);
        $response->assertJsonMissing(["cabin" => "PremiumFirst"]);
    }

    protected function assertReceivedCompletePricingDetailsForSelectedOffers(TestResponse $response): void
    {
        $response->assertOk();
        $response->assertJsonFragment(["@type" => "OfferListResponse"]);
        $response->assertJsonFragment(["@type" => "Offer"]);
        $response->assertJsonFragment(["@type" => "ProductAir"]);
        $response->assertJsonFragment(["@type" => "FlightSegment",]);
    }

    protected function assertWorkbenchCreatedAndReservationIdReceived(TestResponse $response): void
    {
        $response->assertOk();
        $response->assertJsonStructure([
            "ReservationResponse" => [
                "@type",
                "Reservation" => [
                    "@type",
                    "Identifier" => [
                        "authority",
                        "value"
                    ]
                ]
            ]
        ]);
        $response->assertJsonFragment(["@type" => "ReservationResponse"]);
        $response->assertJsonFragment(["@type" => "Reservation"]);
        $this->assertIsString($response->json('ReservationResponse.Reservation.Identifier.value'));
    }

    protected function assertOfferAddedToWorkbenchAndOfferIdReceived(TestResponse $response): void
    {
        $response->assertOk();
        $response->assertJsonStructure([
            "OfferListResponse" => [
                "@type",
                "OfferID" => [
                    "*" => [
                        "@type",
                        "Identifier" => [
                            "authority",
                            "value"
                        ]
                    ]
                ]
            ]
        ]);
        $response->assertJsonFragment(["@type" => "OfferListResponse"]);
        $response->assertJsonFragment(["@type" => "OfferID"]);
        $this->assertIsString($response->json('OfferListResponse.OfferID.0.Identifier.value'));
    }

    protected function assertTravelerAddedWorkbenchAndTravelerIdReceived(TestResponse $response): void
    {
        $response->assertOk();
        $response->assertJsonStructure([
            "TravelerResponse" => [
                "@type",
                "Traveler" => [
                    "@type",
                    "Identifier" => [
                        "value"
                    ]
                ]
            ]
        ]);
        $response->assertJsonFragment(["@type" => "TravelerResponse"]);
        $response->assertJsonFragment(["@type" => "TravelerIdentifier"]);
        $this->assertIsString($response->json('TravelerResponse.Traveler.Identifier.value'));
    }

    protected function assertReservationCommitted(TestResponse $response): void
    {
        $response->assertOk();
        $response->assertJsonFragment(["@type" => "ReservationResponse"]);
        $response->assertJsonFragment(["@type" => "Reservation"]);
        $response->assertJsonFragment(["@type" => "Offer"]);
        $response->assertJsonFragment(["@type" => "Traveler"]);
        $response->assertJsonFragment(["@type" => "ReceiptConfirmation"]);
//        $response->assertJsonFragment(["@type" => "ReservationComment"]);
        $response->assertJsonFragment(["authority" => "Travelport"]);
    }

    protected function assertReceivedOnlyNonStopFlightOffers(TestResponse $response): void
    {
        $this->assertFlightOffersReceivedWithRequiredStructure($response);
        $response->assertJsonFragment(["numberOfStops" => 0]);
        $response->assertJsonMissing(["numberOfStops" => 1]);
        $response->assertJsonMissing(["numberOfStops" => 2]);
        $response->assertJsonMissing(["numberOfStops" => 3]);
    }
}
